#include <bios/types.h>
#include <bios/dec21285.h>
#include <bios/pci.h>
#include <bios/machine.h>
#include <bios/malloc.h>
#include <bios/stdio.h>
#include <bios/string.h>

static unsigned long pci_io_free;
static unsigned long pci_mem_free;

static struct pci_dev *root_dev;

#ifndef DEBUG
#define debug_printf(x...)
#endif

#define FUNC(x) ((x) & 7)
#define SLOT(x) ((x) >> 3)
#define MAX_DEV (20 << 3)

static inline unsigned long pci_alloc(unsigned long *resource, unsigned long size)
{
	unsigned long base;

	*resource += size - 1;
	*resource &= ~(size - 1);

	base = *resource;

	*resource += size;

	return base;
}

static unsigned long pci_alloc_io(unsigned long size)
{
	return pci_alloc(&pci_io_free, size);	
}

static unsigned long pci_alloc_mem(unsigned long size)
{
	return pci_alloc(&pci_mem_free, size);
}

struct pci_dev *
pci_lookupclass(struct pci_dev *dev, unsigned short class)
{
	if (!dev)
		dev = root_dev;

	while (dev) {
		if (dev->class == class &&
		    dev->command & (PCI_COMMAND_MEM_SPACE|PCI_COMMAND_IO_SPACE))
			break;
		dev = dev->next;
	}

	return dev;
}

struct pci_dev *
pci_lookup_vendor_device(struct pci_dev *dev,
			 unsigned short vendor_id, unsigned short device_id)
{
	if (!dev)
		dev = root_dev;

	while (dev) {
		if (dev->vendor == vendor_id &&
		    dev->device == device_id)
			break;
		dev = dev->next;
	}

	return dev;
}

void pci_set_master(struct pci_dev *dev)
{
	dev->command |= PCI_COMMAND_BUS_MASTER;

	pci_write_config_word(dev->pci_base, PCI_COMMAND, dev->command);
}

static void pci_init_hw(void)
{
	unsigned int ctrl = csr_read_long(CSR_CTRL);

	if (ctrl & (1 << 31)) {
		csr_write_long(0, CSR_PCIEXTN);

		/*
		 * Set up the mask registers
		 */
		csr_write_long(12, CSR_OIMR);
		csr_write_long(0, CSR_DPMR);
		csr_write_long(0, CSR_DSMR);

		/*
		 * Disable PCI reset
		 */
		ctrl |= (1 << 9);
		csr_write_long(ctrl, CSR_CTRL);

		/*
		 * Disable our PCI interface
		 */
		csr_write_word(0, CSR_PCI_CMD);

		/*
		 * Set up the PCI -> Host mappings
		 */
		csr_write_long(0, CSR_CSRBASEADDRMASK);
		csr_write_long(0x40000000, CSR_CSRMEMBASE);
		csr_write_long(0x0000f000, CSR_CSRIOBASE);

		/*
		 * SDRAM is mapped at PCI space address 0x10000000
		 */
		csr_write_long(0x01fc0000, CSR_SDRAMBASEADDRMASK);
		csr_write_long(0x10000000, CSR_SDRAMBASE);
		csr_write_long(0, CSR_SDRAMBASEOFF);
		csr_write_word((0x17 | (1 << 9)), CSR_PCI_CMD);

		/*
		 * Set Init Complete
		 */
		csr_write_long(ctrl | CSR_CTRL_INITCOMPLETE, CSR_CTRL);
	}
}

static void pci_scan_bus(void)
{
	struct pci_dev *pdev, **p;
	int dev, multi = 0;

	for (dev = 0; dev < MAX_DEV; dev++) {
		unsigned long pci_base;
		u32 vendor, device;

		if (FUNC(dev) && !multi)
			continue;

		pci_base = pci_config_addr(SLOT(dev), FUNC(dev));

		vendor = pci_read_config_word(pci_base, PCI_VENDOR_ID);
		device = pci_read_config_word(pci_base, PCI_DEVICE_ID);

		if (vendor == 0xffff || vendor == 0x0000 ||
		    device == 0xffff || device == 0x0000)
			continue;

		pdev = malloc(sizeof(*pdev));
		memset(pdev, 0, sizeof(*pdev));
		pdev->vendor   = vendor;
		pdev->device   = device;
		pdev->pci_base = pci_base;
		pdev->header   = pci_read_config_byte(pci_base, PCI_HEADER_TYPE);
		pdev->command  = pci_read_config_word(pci_base, PCI_COMMAND);
		pdev->class    = pci_read_config_long(pci_base, PCI_CLASS_CODE) >> 16;
		pdev->dev      = dev;

		multi = pdev->header & 0x80;

		/*
		 * Add this device to the list
		 */
		for (p = &root_dev; *p; p = &(*p)->next);
		*p = pdev;
	}
}

static void pci_setup_cards_0(struct pci_dev *dev)
{
	unsigned long pci_base = dev->pci_base;
	unsigned long base, size;
	unsigned int i, class;

	for (i = 0; i < 6; i ++) {
		pci_write_config_long(pci_base, PCI_BASE0 + i * 4, 0xffffffff);

		base = pci_read_config_long(pci_base, PCI_BASE0 + i * 4);

		if (base == 0)
			continue;

		if (base & 1) {	/* IO space */
			size = -(base & ~3);
			size &= 0xffff;

			base = pci_alloc_io(size) | 1;

			dev->command |= PCI_COMMAND_IO_SPACE;
		} else {	/* MEM space */
			size = -(base & ~15);

			base = pci_alloc_mem(size);

			dev->command |= PCI_COMMAND_MEM_SPACE;
		}

		dev->bar[i] = base;

		pci_write_config_long(pci_base, PCI_BASE0 + i * 4, base);
		if (pci_read_config_long(pci_base, PCI_BASE0 + i * 4) != base)
			debug_printf("Unable to set base\n");
	}

	pci_write_config_byte(pci_base, PCI_LATENCY_TIMER, 32);
	pci_write_config_long(pci_base, PCI_BIOSROMCONTROL, -2);

	base = pci_read_config_long(pci_base, PCI_BIOSROMCONTROL);
	if (base) {
		size = -(base & ~1);

		base = pci_alloc_mem(size) | 1;

		pci_write_config_long(pci_base, PCI_BIOSROMCONTROL, base);
		dev->bar[7] = base;

		dev->command |= PCI_COMMAND_MEM_SPACE;
	}

	class = pci_read_config_long(pci_base, PCI_CLASS_CODE) >> 24;
	switch (class) {
	case PCI_CLASS_NOT_DEFINED:
	case PCI_BASE_CLASS_DISPLAY:
	case PCI_BASE_CLASS_STORAGE:
		dev->command |= PCI_COMMAND_IO_SPACE;
		break;
	}
}

static void pci_setup_cards(void)
{
	struct pci_dev *dev;

	for (dev = root_dev; dev; dev = dev->next) {
		dev->command &= ~(PCI_COMMAND_BUS_MASTER |
				  PCI_COMMAND_MEM_SPACE |
				  PCI_COMMAND_IO_SPACE);

		if ((dev->header & 0x7f) == 0)
			pci_setup_cards_0(dev);

		pci_write_config_word(dev->pci_base, PCI_COMMAND, dev->command);
	}
}

void pci_print_config(void)
{
	struct pci_dev *dev;

	printf("---------------------- PCI Configuration -----------------------\n");
	printf("Slot Vendor Device Class BMI Address0 Address1 Address2 Address3 Address4 Address5 Bios\n");

	for (dev = root_dev; dev; dev = dev->next) {
		int i;

		printf("%2d.%d", SLOT(dev->dev), FUNC(dev->dev));
		printf("  %04X   %04X  %04X  %c%c%c",
		       dev->vendor, dev->device, dev->class,
		       dev->command & PCI_COMMAND_BUS_MASTER ? '*' : ' ',
		       dev->command & PCI_COMMAND_MEM_SPACE  ? '*' : ' ',
		       dev->command & PCI_COMMAND_IO_SPACE   ? '*' : ' ');

		for (i = 0; i < 7; i++)
			if (dev->bar[i])
				printf(" %08x", dev->bar[i]);
			else
				printf("         ");
		printf("\n");
	}

	printf("----------------------------------------------------------------\n");
}

void pci_init(void)
{
	pci_io_free  = 0x6000;
	pci_mem_free = PCIMEM_BASE;

	pci_init_hw();
	pci_scan_bus();
	pci_setup_cards();
	pci_print_config();
}
